// Source code for BlasterWorks

// Use for the blasters
// NOTE: Values for my_number must be UNIQUE to each blaster. 
// Written by Max Rothstein and Richard Speranza
// for ECE 4760 at Cornell University
// Date: 12-16-2013

// --- Includes and Defines ---------------------------------------
#include "trtSettings.h"
#include "trtkernel644.c"
#include <stdio.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "lcd_lib.h"
#include "lcd_lib.c"
#include "wireless.c"
#include "wireless.h"

#define PIPE_ADDRESS_0  0xb4b5b600  // wireless address
#define SEM_SHARED      1           // semaphore to protect shared variable
#define F_CPU           16000000UL  // Set CPU clock to 16 MHz

// default game values
#define MAX_AMMO        10
#define MAX_HEALTH      100
#define NUM_PLAYERS     2

// wireless state constants
#define INIT_RX     0
#define POLL_UPDATE 1
#define INIT_TX     2
#define SEND_UPDATE 3
#define START_GAME  4

// lcd state constants
#define SELF_INFO       0
#define BLASTER_TYPE    1
#define LEADERBOARD     2

// blaster types
#define SINGLE_SHOT	0
#define AUTO	    1

// --- Function Prototypes -----------------------------------
// initialization
void initialize(void);
void lcd_initialize(void);
// LED toggle
void GRN_toggle(void);
void RED_toggle(void);
// gameplay buttons
char trigger_pressed(void);
char reload(void);
char lcd_cycle(void);
char switch_blaster_type(void);
// TRT tasks
void task_sound_effects(void*);
void task_wireless(void*);
void task_laser(void*);

// --- Global Variables --------------------------------------
int my_ammo[2];
char my_number = 0; // player id 
char blaster_type;
int my_health;
int health[2];
int ammo[2];
int max_health = MAX_HEALTH;
int max_ammo[2] = {MAX_AMMO,MAX_AMMO};

volatile char trigger_interrupt = 0;
volatile char blaster_disable = 0;
volatile char strong_hit = 0;
volatile char weak_hit = 0;

const int8_t LCD_initialize[] PROGMEM = "LCD Init\0";
int8_t lcd_buffer1[17];	// LCD display buffer
int8_t lcd_buffer2[17];	// LCD display buffer
int args[2] ;

char data[] = "           ";

// ---- Interrupts -------------------------------------------
// interrupt on laser hit
ISR ( INT1_vect ) {
    char outer_shot; 
    char blaster_shot; 
    // blaster shot indicator is wired to pinA.6
    blaster_shot    = (PINA & 0x40) == 0x40; 
    // outer shot indicator is wired to pinA.5
    outer_shot      = (PINA & 0x20) == 0x20;

    if ( outer_shot == 1 || blaster_shot == 1 ) weak_hit = 1;
    else strong_hit = 1;
    
    // blaster is disabled when target at the blaster is hit
    if ( blaster_shot == 1 ) blaster_disable = 1;
}

// interrupt on blaster fire request
ISR ( INT2_vect ) {
	trigger_interrupt = 1;
}

// ---- Tasks ------------------------------------------
// updates player health 
void task_health(void* args) {
    long release, deadline;
    while ( 1 ) {
        trtWait(SEM_SHARED);
        // center hit causes 10 points of damage, other hits
        // cause 5 points of damage
        if ( strong_hit == 1 ) my_health -= 10;
        else if ( weak_hit == 1 ) my_health -=5;
        trtSignal(SEM_SHARED);
        
        // players are allowed 3 seconds of immunity after getting hit
        if ( strong_hit == 1 || weak_hit == 1 ) {
            release = trtCurrentTime() + SECONDS2TICKS(3);
            deadline = release + SECONDS2TICKS(0.050);
        }
        else {
            release = trtCurrentTime() + SECONDS2TICKS(0.010);
            deadline = release + SECONDS2TICKS(0.010);
        }
        trtSleepUntil(release,deadline);
        strong_hit = 0;
        weak_hit = 0;
    }
}

// fires the laser blaster
// the gate to the MOSFET which activates the laser is connected to portA.7
void task_laser(void* args) {
	long release, deadline;
	unsigned char laser_on = 0; // enforces on/off behavior of laser
	blaster_type = SINGLE_SHOT; // start out using single shot
	blaster_disable = 0;
	trigger_interrupt = 0;
	
	while ( 1 ) {
        // protect my_ammo and blaster_type
        trtWait(SEM_SHARED);
		if ( blaster_disable == 0 && my_ammo[blaster_type] > 0 ) {
			// auto blaster type fires when trigger is pressed while the single
            // shot mode responds to the external interrupt
            if ( blaster_type == AUTO ) {
				if ( laser_on == 1 ) {
					laser_on = 0;
					PORTA &= 0x7F; // turn port off
				}
				else if ( trigger_pressed() == 1 ) {
					laser_on = 1; 
					PORTA |= 0x80; // turn port on
					my_ammo[blaster_type] -= 1; // decrement ammo	
				}
				release = trtCurrentTime() + SECONDS2TICKS(0.25);
				deadline = release + SECONDS2TICKS(0.05);
			}
			else if ( blaster_type == SINGLE_SHOT ) { 
				if ( laser_on == 1 ) {
					laser_on = 0;
					PORTA &= 0x7F; // turn port off
                    trigger_interrupt = 0;
				}
				else if ( trigger_interrupt == 1 ) {
					laser_on = 1;
					PORTA |= 0x80;  // turn port on
					my_ammo[blaster_type] -= 1; // decrement ammo	
				}
				release 	= trtCurrentTime() + SECONDS2TICKS(0.25);
				deadline 	= release + SECONDS2TICKS(0.05);
			}
		}
		else {
            PORTA &= 0x7F; // disable blaster
			release		= trtCurrentTime() + SECONDS2TICKS(5); // sleep for 5 seconds
			deadline 	= release + SECONDS2TICKS(0.5);
			blaster_disable = 0;
		}
		trtSignal(SEM_SHARED);
		trtSleepUntil(release,deadline);
	}
}

// State machine for wireless communications
void task_wireless(void* args) {
    long release, deadline;
	PAYLOAD_LENGTH = 32; //set the payload length of each packet
    char w_state = INIT_RX;
    w_state = START_GAME;
    // timeout variables
    char first_entry = 1;
    long enter_time = trtCurrentTime();
    int my_prev_health = -1;
    char command; // either S for startup or U for update
    
	init_RF(0); //disable enhanced shockburst
	char* packet = (char *) malloc(PAYLOAD_LENGTH);
	set_mode(RX_MODE);
    set_RX_addr(PIPE_ADDRESS_0, 0);
    initialize();
	while (1) {
        // begin by waiting for initialization packet from base
        if ( w_state == START_GAME ) {
            RED_toggle();
            if (RX_FIFO_status() == 1) {
                read_RX_payload(packet);
                trtWait(SEM_SHARED);
                sscanf(packet,"%c %d %d %d",&command,&max_health,
                    &max_ammo[0],&max_ammo[1]);
                if ( command == 'S' ) {
                    // initialize game state
                    my_health           = max_health;
                    for ( int i = 0; i < NUM_PLAYERS; i++ ) {
                        health[i]       = max_health;
                        ammo[i]         = max_ammo[0];
                        my_ammo[i]      = max_ammo[i];
                    }
                    w_state = INIT_RX;
                }
                trtSignal(SEM_SHARED);
            }
        } 
        // set wireless to receive mode
        else if ( w_state == INIT_RX ) {
            flush_RX_FIFO();
            set_mode(RX_MODE);
            w_state = POLL_UPDATE;
            first_entry = 1;
        }
        // wait on updates from base station or other blasters
        else if ( w_state == POLL_UPDATE ) {
            if (RX_FIFO_status() == 1) {
                read_RX_payload(packet);
               
                int ammo_update[2];
				int health_update[2];
				sscanf(packet,"%c %d %d %d %d",&command,&health_update[0],&ammo_update[0],&health_update[1],&ammo_update[1]);
				
                trtWait(SEM_SHARED);
				if ( command == 'U' ) {
                    for ( int i = 0 ; i < NUM_PLAYERS ; i++ ) {
                        if ( health_update[i] < health[i] ) {
                            health[i] = health_update[i];
                        }
                        if ( ammo_update[i] != -1 && ammo_update[i] < ammo[i] ) {
                            ammo[i] = ammo_update[i];
                        }
                    }
                }
                trtSignal(SEM_SHARED);
            }
            trtWait(SEM_SHARED);
            if ( my_prev_health != my_health ) { 
                // update info on system about life/etc
                if ( my_number == 0 ) {
                    sprintf(data,"U %d %d %d %d",my_health,my_ammo[0],health[1],ammo[1]);
                }
                else {
                    sprintf(data,"U %d %d %d %d",health[0],ammo[1],my_health,my_ammo[0]);
                }

                w_state = INIT_TX;
                my_prev_health = my_health;
            }
            trtSignal(SEM_SHARED);
        }
        // send update to base station
        else if ( w_state == INIT_TX ) {
            flush_RX_FIFO();
            set_mode(TX_MODE);
            set_TX_addr(PIPE_ADDRESS_0);
            w_state = SEND_UPDATE;
        }
        else if ( w_state == SEND_UPDATE ) {
            // capture time when first entering the state
            if(first_entry == 1) {
                first_entry = 0;
                enter_time = trtCurrentTime();
            }
            
            strncpy(packet, data, PAYLOAD_LENGTH); // copy data to sized packet
            transmit_packet(packet);
            
            // timeout after 1 seconds of waiting for the challenge
            if( (trtCurrentTime() - enter_time) > SECONDS2TICKS(1)) {
                w_state = INIT_RX;
                for ( int i = 0; i < NUM_PLAYERS; i++ ) {
                    // if any player reaches 0 or below, the game ends since 
                    // there are only two players, return to startup state
                    // and await reinitialization
                    if ( health[i] <= 0 ) {
                        w_state = START_GAME;
                        flush_RX_FIFO();
                        set_mode(RX_MODE);
                        break;
                    }
                }
            }
        }
        
        // shorter sleep duration while attempting to send update
        if ( w_state == SEND_UPDATE ) {
            release 	= trtCurrentTime() + SECONDS2TICKS(.010);
            deadline 	= release;
        }
        else { 
            release 	= trtCurrentTime() + SECONDS2TICKS(.100);
            deadline 	= release + SECONDS2TICKS(.050);     
        }
        trtSleepUntil(release,deadline);
        
	}
}

// periodically updates the LCD screen with game state
void task_lcd_update(void* args) {
    long release,deadline;
    
    // local status variables
    int local_health[NUM_PLAYERS];
    int prev_health[NUM_PLAYERS];
    int local_ammo[NUM_PLAYERS];
    int i;
    
    // state variables
    char lcd_state      = SELF_INFO;
    char prev_lcd_state = SELF_INFO;
    
    int my_local_health = MAX_HEALTH;
    int my_local_ammo   = MAX_AMMO;
    
    int my_prev_ammo        = MAX_AMMO;
    int my_prev_health      = MAX_HEALTH;
    char local_blaster_type;
    char prev_blaster_type;
    prev_health[0]      = MAX_HEALTH;
    prev_health[1]      = MAX_HEALTH;   
    
    char change_occured = 0; // change in any scoreboard values
   
	lcd_initialize(); 
    
	while (1) {
        // protect values w/ semaphores, copy into local version
        // so display doesn't hog cpu
		trtWait(SEM_SHARED);
        if ( switch_blaster_type() == 1 ) {
			if ( blaster_type == AUTO ) blaster_type = SINGLE_SHOT;
            else if ( blaster_type == SINGLE_SHOT ) blaster_type = AUTO;
		}
        // only allow reloads on SINGLE SHOT
        if ( reload() == 1 && blaster_type == SINGLE_SHOT ) {
            my_ammo[SINGLE_SHOT] = max_ammo[SINGLE_SHOT];
        }
        for ( i = 0; i < NUM_PLAYERS; i++ ) {
            local_health[i] = health[i];
            local_ammo[i]   = ammo[i];
        }
        
        my_local_health = my_health;
        local_blaster_type = blaster_type;
        my_local_ammo   = my_ammo[blaster_type];
        trtSignal(SEM_SHARED);
        
        // change display in response to button push
        if ( lcd_cycle() == 1 ) { 
            lcd_state += 1;
            if ( lcd_state > LEADERBOARD ) {
                lcd_state = SELF_INFO;
            }
        }
        
        // choose between possible display types
        if ( lcd_state == SELF_INFO ) {
            sprintf(lcd_buffer1,"HP: %d", my_local_health); 
            sprintf(lcd_buffer2,"AM: %d", my_local_ammo); 
		}
        else if ( lcd_state == BLASTER_TYPE ) {
            if ( local_blaster_type == SINGLE_SHOT ) {
                sprintf(lcd_buffer1,"SINGLE");
                sprintf(lcd_buffer2,"SHOT");
            }
            else {
                sprintf(lcd_buffer1,"AUTO");
                sprintf(lcd_buffer2,"   ");
            }
            // check to see if display value changed
            if ( prev_blaster_type != local_blaster_type ) {
                change_occured = 1;
            }
        }
        else if ( lcd_state == LEADERBOARD ) {
            sprintf(lcd_buffer1,"P1: %d", local_health[0]);
            sprintf(lcd_buffer2,"P2: %d", local_health[1]);
        }
        else {
            sprintf(lcd_buffer1,"undefined LCD state!");
        }
        // check to see if any display values have changed
        for ( i = 0; i < NUM_PLAYERS; i++ ) {
            if ( prev_health[i] != local_health[i] )
                change_occured = 1;
        }
        if ( my_prev_health != my_local_health || my_prev_ammo != my_local_ammo )
            change_occured = 1;
        
        // only update when state is changed
        if ( lcd_state != prev_lcd_state  || change_occured == 1 ) {
            LCDclr();
            LCDGotoXY(0,0);
            LCDstring(lcd_buffer1,strlen(lcd_buffer1)); // print to LCD
            LCDGotoXY(0,1);
            LCDstring(lcd_buffer2,strlen(lcd_buffer2)); // print to LCD
            
            // sleep for a while after screen updates
            release 	= trtCurrentTime() + SECONDS2TICKS(.3);
            deadline 	= release + SECONDS2TICKS(0.5);
            
            change_occured = 0;
        }
        else {
            // sleep for a short duration without update
            release 	= trtCurrentTime() + SECONDS2TICKS(.050);
            deadline 	= release + SECONDS2TICKS(0.5);
        }
        
        // save previous state
        prev_lcd_state  = lcd_state;
        my_prev_ammo       = my_local_ammo;
        my_prev_health     = my_local_health;
        prev_blaster_type = local_blaster_type;
        
        for ( i = 0; i < NUM_PLAYERS; i++ )
            prev_health[i] = local_health[i];
            
		trtSleepUntil(release,deadline);
	}
}

// --- Main Program ----------------------------------
int main(void) {
	// start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack

	// variable protection
	trtCreateSemaphore(SEM_SHARED, 1) ; // protect shared variables

	// --- create tasks  ----------------
	trtCreateTask(task_wireless, 1000, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[1]));
    trtCreateTask(task_lcd_update, 1000, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[1]));
	trtCreateTask(task_laser, 300, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[1]));
    trtCreateTask(task_health, 300, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[1]));

	// --- Idle task --------------------------------------
	while (1);

} // main

// --- Other Functions ---------------------------------------
void initialize(void) {
    // LED and button initialization   
    DDRA = ( 1 << PINA2 ) | ( 1 << PINA3 ) | ( 1 << PINA7);
    PORTA = 0b10011;  // button pull-up
	PORTB |= ( 1 << PINB2 ); // laser button pull-up
	PORTD |= 0b10000000; // blaster type switch pull-up
    
    // EXTERNAL INTERRUPT CONFIG
    EIMSK 	= (1 << INT1) | (1 << INT2); // enable ext. interrupt on D.3 and B.2
    // set for rising edge interrupt 1, falling-edge interrupt 2
    EICRA 	= (1 << ISC10)|(1 << ISC11) | (1 << ISC21); 
}
 // initialize the lcd screen
void lcd_initialize(void) {
	LCDinit();	//initialize the display
	LCDcursorOFF();
	LCDclr();	//clear the display
	LCDGotoXY(0,0);
	CopyStringtoLCD(LCD_initialize, 0, 0);
}
//button push on pina.0
char reload(void) {
    return ( PINA & 0b1 ) != 0b1;
}
// button push on pina.1
char trigger_pressed(void) {
    return (PINA & 0b10) != 0b10;
}
// button push on pina.4
char lcd_cycle(void) {
    return (PINA & 0b10000) != 0b10000;
}
// button push on pind.7
char switch_blaster_type(void) {
    return (PIND & 0b10000000) != 0b10000000;
}
// toggle the green LED
void GRN_toggle(void) {
	PORTA = PORTA ^ 0x04; 
}
// toggle the red LED
void RED_toggle(void) {
	PORTA = PORTA ^ 0x08; 
}